"""
Validation Functions that help mesh relations
"""

from faux_sures.not_db import Model
from faux_sures.exceptions import OneToOneException


def one_to_one(from_model: Model, from_field: str, to_model: Model, to_field: str) -> bool:
    """
    Create a validator that enforces a One to One field

    Args:
        from_model (Model): First model
        from_field (str): First models field
        to_model (Model): Second model
        to_field (str): Second models field

    Returns:
        True if both references point back to each other
    """

    if not getattr(from_model, from_field) == to_model:
        raise OneToOneException(f"{from_model} to {to_model} is not One to One")
    if not getattr(to_model, to_field) == from_model:
        raise OneToOneException(f"{from_model} to {to_model} is not One to One")
    return True



import sys


class REPLException(Exception):
    """Suppress traceback to maintain flow of Ipython"""

    __suppress_context__ = True


class ModelNotFoundError(REPLException):
    pass


class UniqueFieldRequirementException(REPLException):
    pass


class TypeFieldRequirementException(REPLException):
    pass


class ValidatorFieldRequirementException(REPLException):
    pass


class OneToOneException(REPLException):
    pass


def repl_hook(kind, message, traceback):
    if REPLException in kind.__bases__:
        print("{0}: {1}".format(kind.__name__, message))  # Only print Error Type and Message
    else:
        sys.__excepthook__(kind, message, traceback)  # Print Error Type, Message and Traceback


sys.excepthook = repl_hook



"""
Classes for modeling data 
"""
from __future__ import annotations

import functools
import pydoc
import uuid

from collections import defaultdict
from dataclasses import dataclass, field as dataclass_field, fields as list_dataclass_fields
from typing import Any, Callable, Dict, Iterable, Optional, Type, Tuple, Union, Set, Generator, no_type_check

from faux_sures.exceptions import (
    ModelNotFoundError,
    TypeFieldRequirementException,
    UniqueFieldRequirementException,
    ValidatorFieldRequirementException,
)

ValidatorType = Callable[[Any], bool]


@no_type_check
@functools.lru_cache
def get_type_from_class_name(class_name: Union[str, Type]) -> Type:
    if isinstance(class_name, Type):
        return class_name
    return pydoc.locate(class_name)


class Field:
    """
    A field for a model complete with typing and validation
    """

    def __init__(
        self,
        _type: Union[Type, str],
        validators: Union[ValidatorType, Iterable[ValidatorType]] = tuple(),
        optional: bool = False,
    ):
        """
        Args:
            _type (Type or Str): Appropriate Type. If string instead of type object is passed, lexical cast to type at set time
            validators (Iterable[Callable[[Any] Any]]): Single or Collection of validation functions
            optional (bool): Whether this value is allowed to be None and skip the type and validation checks
        """

        if isinstance(_type, str):
            self.lexical_cast = True
        else:
            self.lexical_cast = False
        self._type = _type
        self.validators: Tuple[ValidatorType] = (
            tuple(validators) if isinstance(validators, Iterable) else (validators,)  # type: ignore
        )
        self.optional = optional
        self.name: Optional[str]

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if not instance:
            return self
        return instance.__dict__.get(self.name, None)

    def __delete__(self, instance):
        del instance.__dict__[self.name]

    def __set__(self, instance, value):

        if self.lexical_cast:
            self._type = get_type_from_class_name(self._type)
            self.lexical_cast = False

        if value is None and self.optional is True:
            pass
        elif not isinstance(value, self._type):
            raise TypeFieldRequirementException(
                f"{self.name!r} values must one of types {self._type!r} not {type(value)}"
            )
        if value is not None:
            for validator in self.validators:
                if validator(value) is False:
                    raise ValidatorFieldRequirementException(f"{self.name} failed to validate {validator.__name__}")
        instance.__dict__[self.name] = value


class UniqueTogetherRestraint:
    """
    Add a check for unique values over a single or a series of values
    """

    def __init__(self, unique_fields: Union[str, Iterable[str]]):
        if not unique_fields:
            raise ValueError("Fields must be supplied for unique constraint")
        self.unique_fields: Tuple[str] = (
            (unique_fields,) if isinstance(unique_fields, str) else tuple(unique_fields)  # type: ignore
        )


@dataclass
class Model:
    """
    Base Data Model
    """

    pk: str = dataclass_field(default_factory=lambda: str(uuid.uuid4()))

    @property
    def class_name(self):
        return str(self.__class__.__name__)

    def __deepcopy__(self, memodict={}):
        """Never duplicate internal key. Preserve unset Fields"""
        _dcopy = self.__class__()
        fields = {f.name for f in list_dataclass_fields(self)}
        fields.remove("pk")
        for _field, value in self.__dict__.items():
            if _field in fields:
                setattr(_dcopy, _field, value)
        return _dcopy

    def serialize(self) -> Dict[str, str]:
        """
        Return dictionary representation of self
        None is represented as empty string
        """
        fields = {f.name for f in list_dataclass_fields(self)}
        return {f: v if v is not None else "" for f, v in self.__dict__.items() if f in fields}

    def checks(self):
        """Run more complicated check functions over the fields of single model"""
        pass

    def iterate_model_fields(self) -> Generator[Optional[str], None, None]:
        """
        Iterate over all attributes of type Field defined for the class
        """
        for _field in vars(self.__class__).values():
            if isinstance(_field, Field):
                yield _field.name

    def iterate_unique_constraints(self) -> Generator[Tuple, None, None]:
        """
        Iterate over all unique constraints and yield values together

        Raises Attribute Exception if field does not exist on Model
        """
        unique_constraints = (
            c_field for c_field in vars(self.__class__).values() if isinstance(c_field, UniqueTogetherRestraint)
        )
        for constraint in unique_constraints:
            values = (getattr(self, value) for value in constraint.unique_fields)
            yield (self.class_name, *values)

    def validate(self):
        """
        Run all validation for a single model
        Raises
            TypeFieldRequirementException if a field is set to None when it is not Optional
            ValidatorFieldRequirementException if a field does not pass a set Validation Requirement
            Other Custom Exceptions if defined on the model


        Args:
            model (Model): Model with Fields for validation and custom checks
        """

        for _field_name in self.iterate_model_fields():
            if getattr(self, _field_name) is None:
                setattr(self, _field_name, None)  # trigger __set__ checks
        self.checks()


class Session:
    """
    Create a singleton session that stores the models and runs validation at commit
    """

    class Indexes:
        """
        An extendable collection of indexes of references
        """

        model_index: Dict[str, Model] = dict()
        class_index: Dict[str, Set[str]] = defaultdict(set)

        @classmethod
        def flush(cls):
            """
            Clear contents and restart
            """
            cls.model_index = dict()
            cls.class_index = defaultdict(set)

    @classmethod
    def add(cls, model: Model):
        """Add a new Model to session and update references"""

        key = model.pk
        cls.Indexes.model_index[key] = model
        cls.Indexes.class_index[model.class_name].add(key)

    @classmethod
    def query_by_key(cls, pk: str) -> Model:
        """From a key recover the original Model. Raises ModelNotFoundError if not found"""
        try:
            return cls.Indexes.model_index[pk]
        except KeyError:
            raise ModelNotFoundError(f"No record {pk} in session")

    @classmethod
    def query_by_type_name(cls, class_name: str) -> Generator[Model, None, None]:
        """Iterate over all objects of a class. Raises ModelNotFoundError if class name is not found"""

        try:
            models = cls.Indexes.class_index[class_name]
        except KeyError:
            raise ModelNotFoundError(f"No record of type {class_name} in session")

        for key in models:
            yield cls.query_by_key(key)

    @classmethod
    def remove(cls, pk: str) -> bool:
        """
        Pass a Model key and remove the Model's references from the session

        Args:
            pk (str): A Model's internal key

        Returns:
            bool: If references were removed
        """
        try:
            model = Session.query_by_key(pk)
        except KeyError:
            return False
        del cls.Indexes.model_index[pk]
        cls.Indexes.class_index[model.class_name].remove(pk)
        return True

    @classmethod
    def commit(cls):
        """
        Run validation and index checks over all added models

        Raises:
            TypeFieldRequirementException if a field is set to None when it is not Optional
            ValidatorFieldRequirementException if a field does not pass a set Validation Requirement
            UniqueFieldRequirementException if fields across Models do not match Uniqueness Requirements
            Other Custom Exceptions if defined on the model


        """
        uniqueness_tracker: Set[Tuple] = set()

        def check_unique(values: Tuple) -> bool:
            """Update unique tracker and return True on unique add"""
            if values in uniqueness_tracker:
                return False
            uniqueness_tracker.add(values)
            return True

        model: Model
        for model in cls.Indexes.model_index.values():
            model.validate()
            for unique_tuple in model.iterate_unique_constraints():
                if not check_unique(unique_tuple):
                    raise UniqueFieldRequirementException(
                        f"{model.class_name}: {model.pk} fails unique constraint for {unique_tuple[1:]}"
                    )

    @classmethod
    def reset(cls):
        """
        Clear contents and restart
        """
        cls.Indexes.flush()

    def __enter__(self):
        return self

    def __close__(self, *args):
        self.reset()



"""
Create a 5 person team intramural high school basketball league
and test interactions

In this league, a team is made of players and coaches.
Players and coaches can belong to only one team.

In this league, a player must be between 14 and 19 years old.
In this league, a coach must be at least 25.

In this league, a team must have exactly 5 main players
and up to 4 substitions.
(Generally speaking, a main player commits to at least 80% of games but that's not tracked here... yet)

In this league, each team has exactly 1 captain who is a main player.

In this league, each player must have a unique combination of first and last name across the whole league.
In this league, each player must have a unique number within a team.
In this league, a team must have a unique name less than 50 characters but more than 5.
"""

import pytest

from faux_sures import not_db
from faux_sures.exceptions import OneToOneException, TypeFieldRequirementException, ValidatorFieldRequirementException
from faux_sures.recipes.curries import in_range, max_length, min_length
from faux_sures.recipes.sauces import one_to_one

CURRENT_SCOPE = "example_tests.test_team_sports.test_team_sports"


def class_name(local_class: str):
    return f"{CURRENT_SCOPE}.{local_class}"


class LeagueRulesException(Exception):
    pass


class LeagueRules:
    @staticmethod
    def is_captain(player) -> bool:
        return player.is_captain or False

    @staticmethod
    def player_captain_is_main(player: "Player") -> bool:
        if player.is_captain and not player.is_main:
            raise LeagueRulesException(f"Player {player.first_name + player.last_name} must be a main to be captain")

    @staticmethod
    def unique_player_number_within_team(team: "Team") -> bool:
        """Return True if all players within a team have a unique number on their jersey"""
        all_players = not_db.Session.query_by_type_name("Player")
        team_players = [player for player in all_players if player.team == team]
        unique_jerseys = {player.jersey_number for player in team_players}
        if not len(team_players) == len(unique_jerseys):
            raise LeagueRulesException(f"Not all Players on {team.team_name} have a unique number")
        return True

    @staticmethod
    def only_one_captain_per_team(team: "Team") -> bool:
        """Return false if a team has more than one player who is a captain"""
        all_players = not_db.Session.query_by_type_name("Player")
        team_captain = [player for player in all_players if player.team == team and player.is_captain]
        if not len(team_captain) == 1:
            raise LeagueRulesException(f"{team.team_name} has {team_captain} captains")
        return True

    @staticmethod
    def correct_number_of_players(team: "Team") -> bool:
        """True if exactly 5 main players and less than 5 subs"""
        player_types = {True: 0, False: 0}
        all_players = not_db.Session.query_by_type_name("Player")
        for player in all_players:
            player_types[player.is_main] += 1
        if player_types[True] != 5 or player_types[False] > 4:
            raise LeagueRulesException(
                f"{team.team_name} has {player_types[True]} main players and {player_types[False]} subs"
            )
        return True


class Person(not_db.Model):

    first_name = not_db.Field(str)
    last_name = not_db.Field(str)


class Player(Person):
    """A basketball Player"""

    age = not_db.Field(int, in_range(14, 19))
    is_captain = not_db.Field(bool, optional=True)
    is_main = not_db.Field(bool)
    jersey_number = not_db.Field(int, in_range(1, 99))
    team = not_db.Field(class_name("Team"))

    jersey_constraint = not_db.UniqueTogetherRestraint(("first_name", "last_name"))

    def checks(self):
        LeagueRules.player_captain_is_main(self)


class Coach(Person):

    age = not_db.Field(int, in_range(25, None))
    team = not_db.Field(class_name("Team"))

    def checks(self):
        one_to_one(self, "team", self.team, "coach")


class Team(not_db.Model):

    team_name = not_db.Field(str, (min_length(5), max_length(50)))
    captain = not_db.Field(class_name("Player"), LeagueRules.is_captain)
    coach = not_db.Field(class_name("Coach"))

    name_constraint = not_db.UniqueTogetherRestraint("team_name")

    def checks(self):
        one_to_one(self, "captain", self.captain, "team")
        one_to_one(self, "coach", self.coach, "team")
        LeagueRules.correct_number_of_players(self)
        LeagueRules.only_one_captain_per_team(self)
        LeagueRules.unique_player_number_within_team(self)


LIST_OF_NAMES = [
    ("Johnny", "Silverhands"),
    ("Cool", "Luke"),
    ("Smellya", "Later"),
    ("Jame", "Bigguy"),
    ("Kobe", "Rip"),
]


def test_type_repl_feedback():
    """Immediately assert error when an incorrect type is set"""

    not_db.Session.reset()

    ace = Player()
    with pytest.raises(TypeFieldRequirementException):
        ace.age = "13"


def test_rule_repl_feedback():
    """Immediately assert error when an invalid value is set"""

    not_db.Session.reset()

    ace = Player()
    with pytest.raises(ValidatorFieldRequirementException, match="age failed to validate in_range_14_to_19"):
        ace.age = 13


def test_rule_repl_retry():
    """When you fall down, get back up again"""

    not_db.Session.reset()

    ace = Player()
    with pytest.raises(ValidatorFieldRequirementException, match="age failed to validate in_range_14_to_19"):
        ace.age = 13
    ace.age = 15
    assert ace.age == 15
    with pytest.raises(ValidatorFieldRequirementException, match="age failed to validate in_range_14_to_19"):
        ace.age = 13
    assert ace.age == 15


def test_team_name():

    not_db.Session.reset()

    wolves = Team()
    with pytest.raises(ValidatorFieldRequirementException, match="max_length"):
        wolves.team_name = """
        In an electrophilic aromatic substitution reaction,
        the hydrogen bonded to the arenium ion carbon that is the site of attack 
        by the incoming electrophile.
        """
    assert wolves.team_name is None
    with pytest.raises(ValidatorFieldRequirementException, match="min_length"):
        wolves.team_name = "wolf"
    assert wolves.team_name is None
    wolves.team_name = "wolves"
    assert wolves.team_name == "wolves"


def test_missing_team():

    not_db.Session.reset()

    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.jersey_number = 12
    with pytest.raises(TypeFieldRequirementException, match="'team' values must one of types"):
        ace.validate()


def test_fix_captain_status():

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12

    with pytest.raises(ValidatorFieldRequirementException, match="is_captain"):
        wolves.captain = ace
    ace.is_captain = True
    wolves.captain = ace
    assert wolves.captain == ace


def test_object_references():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    assert wolves.captain == ace
    assert wolves.captain.team == wolves
    wolves.team_name = "Jaguars"
    assert ace.team.team_name == "Jaguars"
    ace.last_name = "Smomo"
    assert wolves.captain.last_name == "Smomo"

    team_reference = next(not_db.Session.query_by_type_name("Team"))
    assert team_reference.team_name == "Jaguars"


def test_good_single_team():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    other_members = (Player() for _ in range(5))
    for pos, member in enumerate(other_members):
        member.first_name = LIST_OF_NAMES[pos][0]
        member.last_name = LIST_OF_NAMES[pos][1]
        member.age = 17
        member.is_main = pos < 4
        member.team = wolves
        member.jersey_number = 27 + pos ** 2
        not_db.Session.add(member)

    coach = Coach()
    coach.first_name = "Coach"
    coach.last_name = "Mann"
    coach.age = 46
    coach.team = wolves
    wolves.coach = coach
    not_db.Session.add(coach)

    not_db.Session.commit()

    assert len(list(not_db.Session.query_by_type_name("Player"))) == 6
    assert len(list(not_db.Session.query_by_type_name("Team"))) == 1
    assert len(list(not_db.Session.query_by_type_name("Coach"))) == 1


def test_unique_name_fail():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    other_members = (Player() for _ in range(5))
    for pos, member in enumerate(other_members):
        member.first_name = LIST_OF_NAMES[pos][0]
        member.last_name = LIST_OF_NAMES[pos][1]
        member.age = 17
        member.is_main = pos < 4
        member.team = wolves
        member.jersey_number = 27 + pos ** 2
        not_db.Session.add(member)
    member.first_name = "Ace"
    member.last_name = "Ventura"

    coach = Coach()
    coach.first_name = "Coach"
    coach.last_name = "Mann"
    coach.age = 46
    coach.team = wolves
    wolves.coach = coach
    not_db.Session.add(coach)

    with pytest.raises(not_db.UniqueFieldRequirementException):
        not_db.Session.commit()


def test_fail_one_to_one_validation():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    other_members = (Player() for _ in range(5))
    for pos, member in enumerate(other_members):
        member.first_name = LIST_OF_NAMES[pos][0]
        member.last_name = LIST_OF_NAMES[pos][1]
        member.age = 17
        member.is_main = pos < 4
        member.team = wolves
        member.jersey_number = 27 + pos ** 2
        not_db.Session.add(member)

    coach = Coach()
    coach.first_name = "Coach"
    coach.last_name = "Mann"
    coach.age = 46
    coach.team = wolves

    fake_coach = Coach()
    fake_coach.first_name = "Coach"
    fake_coach.last_name = "Flynn"
    fake_coach.age = 32
    fake_coach.team = wolves

    wolves.coach = fake_coach
    not_db.Session.add(coach)
    not_db.Session.add(fake_coach)

    with pytest.raises(OneToOneException):
        not_db.Session.commit()


def test_too_many_main_players():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    other_members = (Player() for _ in range(5))
    for pos, member in enumerate(other_members):
        member.first_name = LIST_OF_NAMES[pos][0]
        member.last_name = LIST_OF_NAMES[pos][1]
        member.age = 17
        member.is_main = True
        member.team = wolves
        member.jersey_number = 27 + pos ** 2
        not_db.Session.add(member)

    coach = Coach()
    coach.first_name = "Coach"
    coach.last_name = "Mann"
    coach.age = 46
    coach.team = wolves
    wolves.coach = coach
    not_db.Session.add(coach)

    with pytest.raises(LeagueRulesException, match="6 main players and 0 subs"):
        not_db.Session.commit()


def test_duplicate_jersey_in_team():

    not_db.Session.reset()

    wolves = Team()
    wolves.team_name = "wolves"
    ace = Player()
    ace.first_name = "Ace"
    ace.last_name = "Ventura"
    ace.age = 17
    ace.is_captain = True
    ace.is_main = True
    ace.team = wolves
    ace.jersey_number = 12
    wolves.captain = ace
    not_db.Session.add(ace)
    not_db.Session.add(wolves)

    other_members = (Player() for _ in range(5))
    for pos, member in enumerate(other_members):
        member.first_name = LIST_OF_NAMES[pos][0]
        member.last_name = LIST_OF_NAMES[pos][1]
        member.age = 17
        member.is_main = True
        member.team = wolves
        member.jersey_number = 27 + pos ** 2
        not_db.Session.add(member)
    member.jersey_number = 12
    member.is_main = False

    coach = Coach()
    coach.first_name = "Coach"
    coach.last_name = "Mann"
    coach.age = 46
    coach.team = wolves
    wolves.coach = coach
    not_db.Session.add(coach)

    with pytest.raises(LeagueRulesException, match="Not all Players on wolves have a unique number"):
        not_db.Session.commit()



"""
Utilities for simple validators
"""

from numbers import Number
from typing import Callable, Optional


def exact_length(length: int) -> Callable[[str], bool]:
    """
    Create an EXACT length validator

    Args:
        length (int): exact length

    Returns:
        Callable[[str], bool]: a function with the name exact_length_{length}
    """

    def v(_field):
        return len(_field) == length

    v.__name__ = f"exact_length_{length}"
    return v


def min_length(length: int) -> Callable[[str], bool]:
    """
    Create a MIN length validator

    Args:
        length (int): MIN length (inclusive)

    Returns:
        Callable[[str], bool]: a function with the name min_length_{length}
    """

    def v(_field):
        return len(_field) >= length

    v.__name__ = f"min_length_{length}"
    return v


def max_length(length: int) -> Callable[[str], bool]:
    """
    Create an MAX length validator (inclusive)

    Args:
        length (int): MAX length

    Returns:
        Callable[[str], bool]: a function with the name exact_length_{length}
    """

    def v(_field):
        return len(_field) <= length

    v.__name__ = f"max_length_{length}"
    return v


def in_range(_min: Optional[Number] = None, _max: Optional[Number] = None) -> Callable[[Number], bool]:
    """
    Creates a number range validator
    * It can work for any number type

    Args:
        _max (Optional[Number]): max value
        _min (Optional[Number]): min value

    Returns:
        Callable[[Number], bool]: a function with the name in_range_{}_to_{}
    """
    if _min is None and _max is None:
        raise ValueError("Either min or max must be supplied")

    def v(_field):
        return (_min if _min is not None else _field) <= _field <= (_max if _max is not None else _field)

    v.__name__ = f"in_range_{_min}_to_{_max}"
    return v



"""
Test curries and sauces
"""
import pytest

from numbers import Number

from faux_sures.recipes.curries import max_length, min_length, exact_length, in_range
from faux_sures.not_db import Field, ValidatorFieldRequirementException


@pytest.mark.parametrize("bound", [0, 1, 10, 50])
@pytest.mark.parametrize("inp", ["", "cat", "stevenson", "This is a long string. Hello World!"])
@pytest.mark.parametrize("_function", ["max_length", "min_length", "exact_length"])
def test_string_curries(bound, inp, _function):
    """Test max, min and exact length"""

    if _function == "min_length":
        result = len(inp) >= bound
        func = min_length
    elif _function == "max_length":
        result = len(inp) <= bound
        func = max_length
    elif _function == "exact_length":
        result = len(inp) == bound
        func = exact_length
    else:
        assert False

    partial = func(bound)
    assert result is partial(inp)
    assert partial.__name__ == f"{_function}_{bound}"

    class Simple:
        string_field = Field(str, func(bound))

    simple = Simple()
    if result is True:
        simple.string_field = inp
        assert simple.string_field == inp
    else:
        with pytest.raises(ValidatorFieldRequirementException, match=f"{_function}_{bound}"):
            simple.string_field = inp
        assert simple.string_field is None


@pytest.mark.parametrize("bounds", [(None, 20), (-30, 0), (12.5, 30), (1 / 3, None)])
@pytest.mark.parametrize("inp", [0, 12, 1 / 3, -50, 15, 3.1415 ** 3])
def test_int_range_curry(bounds, inp):
    """Test in_range function"""

    _min, _max = bounds
    if _min is None:
        result = inp <= _max
    elif _max is None:
        result = inp >= _min
    else:
        result = _min <= inp <= _max

    partial = in_range(*bounds)
    assert result is partial(inp)
    assert partial.__name__ == f"in_range_{_min}_to_{_max}"

    class Simple:
        num_field = Field(Number, in_range(*bounds))

    simple = Simple()
    if result is True:
        simple.num_field = inp
        assert simple.num_field == inp
    else:
        with pytest.raises(ValidatorFieldRequirementException, match=f"in_range_{_min}_to_{_max}"):
            simple.num_field = inp
        assert simple.num_field is None
